"""Add complete-field to Sources

Revision ID: 78fa69083a7c
Revises: 87d4ea0aa91d
Create Date: 2019-10-11 19:06:17.373389

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "78fa69083a7c"
down_revision = "87d4ea0aa91d"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("source", sa.Column("complete", sa.Boolean(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("source", "complete")
    # ### end Alembic commands ###
