"""Base

Revision ID: 0

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "0"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass
